/*
 * Copyright (c) 2021, [Ribose Inc](https://www.ribose.com).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "types.h"
#include "str-utils.h"

const char *
id_str_pair::lookup(const id_str_pair pair[], int id, const char *notfound)
{
    while (pair && pair->str) {
        if (pair->id == id) {
            return pair->str;
        }
        pair++;
    }
    return notfound;
}

int
id_str_pair::lookup(const id_str_pair pair[], const char *str, int notfound)
{
    while (pair && pair->str) {
        if (rnp::str_case_eq(str, pair->str)) {
            return pair->id;
        }
        pair++;
    }
    return notfound;
}

int
id_str_pair::lookup(const id_str_pair pair[], const std::vector<uint8_t> &bytes, int notfound)
{
    while (pair && pair->str) {
        if ((strlen(pair->str) == bytes.size()) &&
            !memcmp(pair->str, bytes.data(), bytes.size())) {
            return pair->id;
        }
        pair++;
    }
    return notfound;
}

int
id_str_pair::lookup(const id_str_pair                 pair[],
                    const std::basic_string<uint8_t> &bytes,
                    int                               notfound)
{
    while (pair && pair->str) {
        if ((strlen(pair->str) == bytes.size()) &&
            !memcmp(pair->str, bytes.data(), bytes.size())) {
            return pair->id;
        }
        pair++;
    }
    return notfound;
}
